# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Version information for PyRETIS.

This file is generated by PyRETIS (``setup_version.py``)
"""
SHORT_VERSION = '3.0.1'
VERSION = '3.0.1'
FULL_VERSION = '3.0.1'
GIT_REVISION = '3135263ee53348bf3aa4bf685622f988dcc881f5'
GIT_VERSION = '3.0.1'
RELEASE = True

if not RELEASE:
    VERSION = GIT_VERSION

MIN_PYTHON_VERSION = (3, 9)
