# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This module just contains some info for PyRETIS.

Here we define the name of the program and some other
relevant info.

This file is generated by PyRETIS (``setup_version.py``)
"""
PROGRAM_NAME = 'PyRETIS'
URL = 'http://www.pyretis.org'
GIT_URL = 'https://gitlab.com/pyretis/pyretis'
CITE = """
[1] A. Lervik, E. Riccardi and T. S. van Erp, J. Comput. Chem., 2017
    doi: https://dx.doi.org/10.1002/jcc.24900
[2] E. Riccardi, A. Lervik, S. Roet, O. Aarøen and T. S. van Erp,
    J. Comput. Chem., 2019, doi: https://dx.doi.org/10.1002/jcc.26112
[3] W. Vervust, D. T. Zhang, A. Ghysels, S. Roet, T. S. van Erp, E. Riccardi,
    J. Comput. Chem. 2024, 1. https://doi.org/10.1002/jcc.27319

"""
LOGO = r"""

  _______                                     _______________
 /   ___ \       _____  ____  ____________   / __        ___/
/_/\ \_/ /_  __  \  _ \/ __/ /_  _/ / ___/  /_/ / // // /
   /  __  / / /  / /_)  _/    / // /\ \        / // // / __
  / /   \ \/ /  / _  \ /__   / // /__\ \   ___/ // // /_/ /
 /_/    _\  /  /_/ |_|___/  /_//_/_____/  /______________/
       /___/
"""
